@@ -2,18 +2,21 @@
 
 #include "types.h"
 
-unsigned char *builtins[]=
+const char *builtins[]=
 {
-	USTR "joerc",
-		USTR "-notite         Don't send terminal initialization and termination strings:\n"
+	"joerc",
+		"-notite         Don't send terminal initialization and termination strings:\n"
 		"-assume_color	Assume terminal has ANSI color support even if termcap/terminfo entry\n"
 		"-assume_256color\n"
 		"-guess_non_utf8	Allow guess of non-UTF-8 file encoding in a UTF-8 locale.\n"
+		"-guess_utf16	Allow guess of UTF-16 encoding\n"
 		"-guess_crlf     Automatically detect MS-DOS files and set -crlf appropriately\n"
 		"-guess_indent	Guess indent character (tab or space).\n"
 		"-menu_above	Position menu/list above prompt when enabled.  Otherwise position\n"
 		"-transpose	Transpose rows with columns in all menus.\n"
 		"-joe_state     Use ~/.joe_state file\n"
+		"-brpaste	When JOE starts, send command to the terminal emulator that\n"
+		"-pastehack	If keyboard input comes in as one block assume it's a mouse\n"
 		"-restore	Restore previous cursor position when files are opened\n"
 		"-search_prompting\n"
 		"-lmsg \\i%k%T%W%I%X %n %m%y%R %M %x\n"
@@ -45,8 +48,9 @@ unsigned char *builtins[]=
 		"mode,\"asis\",rtn	% % Meta chars as-is\n"
 		"mode,\"language\",rtn	V Language\n"
 		"mode,\"picture\",rtn	P picture %Zpicture%\n"
+		"mode,\"type\",rtn		F File type [%Ztype%]\n"
 		"menu,\"more-options\",rtn	  % % More options...\n"
-		":defmenu more-options\n"
+		":defmenu more-options menu,\"root\",rtn\n"
 		"menu,\"^G\",rtn	% % ^G options\n"
 		"menu,\"search\",rtn	% % search options\n"
 		"menu,\"paragraph\",rtn	% % paragraph options\n"
@@ -56,7 +60,7 @@ unsigned char *builtins[]=
 		"menu,\"cursor\",rtn	% % cursor options\n"
 		"menu,\"marking\",rtn	% % marking options\n"
 		"menu,\"tab\",rtn	% % tab/indent options\n"
-		":defmenu indent\n"
+		":defmenu indent menu,\"root\",rtn\n"
 		"mode,\"istep\",rtn,\"1\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 1, Indent character = 32\",rtn	1 Space\n"
 		"mode,\"istep\",rtn,\"2\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 2, Indent character = 32\",rtn	2 Spaces\n"
 		"mode,\"istep\",rtn,\"3\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 3, Indent character = 32\",rtn	3 Spaces\n"
@@ -65,12 +69,12 @@ unsigned char *builtins[]=
 		"mode,\"istep\",rtn,\"8\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 8, Indent character = 32\",rtn	8 Spaces\n"
 		"mode,\"istep\",rtn,\"10\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 10, Indent character = 32\",rtn	0 Ten\n"
 		"mode,\"istep\",rtn,\"1\",rtn,mode,\"indentc\",rtn,\"9\",rtn,msg,\"Indent step = 1, Indent character = 9\",rtn	T Tab\n"
-		":defmenu menu\n"
+		":defmenu menu menu,\"more-options\",rtn\n"
 		"mode,\"menu_explorer\",rtn	% % Menu explorer %Zmenu_explorer%\n"
 		"mode,\"menu_above\",rtn	% % Menu position %Zmenu_above%\n"
 		"mode,\"menu_jump\",rtn	% % Jump into menu %Zmenu_jump%\n"
 		"mode,\"transpose\",rtn	% % Transpose menus %Ztranspose%\n"
-		":defmenu ^G\n"
+		":defmenu ^G menu,\"more-options\",rtn\n"
 		"mode,\"highlighter_context\",rtn	% % ^G uses highlighter context %Zhighlighter_context%\n"
 		"mode,\"single_quoted\",rtn	% % ^G ignores '...' %Zsingle_quoted%\n"
 		"mode,\"no_double_quoted\",rtn	% % ^G no ignore \"...\" %Zno_double_quoted%\n"
@@ -81,24 +85,26 @@ unsigned char *builtins[]=
 		"mode,\"semi_comment\",rtn	% % ^G ignores ;... %Zsemi_comment%\n"
 		"mode,\"tex_comment\",rtn % % ^G ignores %%... %Ztex_comment%\n"
 		"mode,\"text_delimiters\",rtn % % Text delimiters %Ztext_delimiters%\n"
-		":defmenu search\n"
+		":defmenu search menu,\"more-options\",rtn\n"
 		"mode,\"icase\",rtn	% % Case insensitivity %Zicase%\n"
 		"mode,\"wrap\",rtn	% % Search wraps %Zwrap%\n"
 		"mode,\"search_prompting\",rtn	% % Search prompting %Zsearch_prompting%\n"
 		"mode,\"csmode\",rtn	% % Continued search %Zcsmode%\n"
-		":defmenu paragraph\n"
+		":defmenu paragraph menu,\"more-options\",rtn\n"
 		"mode,\"french\",rtn	% % French spacing %Zfrench%\n"
 		"mode,\"flowed\",rtn	% % Flowed text %Zflowed%\n"
 		"mode,\"cpara\",rtn	% % Paragraph indent chars %Zcpara%\n"
-		":defmenu file\n"
+		"mode,\"cnotpara\",rtn	% % Not-paragraph chars %Zcnotpara%\n"
+		":defmenu file menu,\"more-options\",rtn\n"
 		"mode,\"restore\",rtn	% % Restore cursor %Zrestore%\n"
 		"mode,\"guess_crlf\",rtn	% % Auto detect CR-LF %Zguess_crlf%\n"
 		"mode,\"guess_indent\",rtn	% % Guess indent %Zguess_indent%\n"
 		"mode,\"guess_non_utf8\",rtn	% % Guess non-UTF-8 %Zguess_non_utf8%\n"
 		"mode,\"guess_utf8\",rtn	% % Guess UTF-8 %Zguess_utf8%\n"
+		"mode,\"guess_utf16\",rtn	% % Guess UTF-16 %Zguess_utf16%\n"
 		"mode,\"force\",rtn	% % Force last NL %Zforce%\n"
 		"mode,\"nobackup\",rtn	% % No backup %Znobackup%\n"
-		":defmenu global\n"
+		":defmenu global menu,\"more-options\",rtn\n"
 		"mode,\"nolocks\",rtn	% % Disable locks %Znolocks%\n"
 		"mode,\"nobackups\",rtn	% % Disable backups %Znobackups%\n"
 		"mode,\"nodeadjoe\",rtn	% % Disable DEADJOE %Znodeadjoe%\n"
@@ -112,16 +118,16 @@ unsigned char *builtins[]=
 		"mode,\"joe_state\",rtn	% % Joe_state file %Zjoe_state%\n"
 		"mode,\"undo_keep\",rtn	% % No. undo records %Zundo_keep%\n"
 		"mode,\"backpath\",rtn	% % Path to backup files %Zbackpath%\n"
-		":defmenu cursor\n"
+		":defmenu cursor menu,\"more-options\",rtn\n"
 		"mode,\"pg\",rtn	% % No. PgUp/PgDn lines %Zpg%\n"
 		"mode,\"mid\",rtn	C Center on scroll %Zmid%\n"
 		"mode,\"floatmouse\",rtn	% % Click past end %Zfloatmouse%\n"
 		"mode,\"rtbutton\",rtn	% % Right button %Zrtbutton%\n"
-		":defmenu marking\n"
+		":defmenu marking menu,\"more-options\",rtn\n"
 		"mode,\"autoswap\",rtn	% % Autoswap mode %Zautoswap%\n"
 		"mode,\"marking\",rtn	% % Marking %Zmarking%\n"
 		"mode,\"lightoff\",rtn	% % Auto unmask %Zlightoff%\n"
-		":defmenu tab\n"
+		":defmenu tab menu,\"more-options\",rtn\n"
 		"mode,\"smarthome\",rtn	% % Smart home key %Zsmarthome%\n"
 		"mode,\"smartbacks\",rtn	% % Smart backspace %Zsmartbacks%\n"
 		"mode,\"indentfirst\",rtn	% % To indent first %Zindentfirst%\n"
@@ -166,29 +172,36 @@ unsigned char *builtins[]=
 		"}\n"
 		"{Search\n"
 		"\\i   Help Screen    \\|turn off with ^KH    prev. screen ^[,    next screen ^[.     \\i\n"
-		"\\i \\iSpecial search sequences:                                                    \\|\\i \\i\n"
-		"\\i \\i    \\\\^  \\\\$  matches beg./end of line      \\\\?     match any single char       \\|\\i \\i\n"
-		"\\i \\i    \\\\<  \\\\>  matches beg./end of word      \\\\*     match 0 or more chars       \\|\\i \\i\n"
-		"\\i \\i    \\\\c      matches balanced C expression \\\\\\\\     matches a \\\\                 \\|\\i \\i\n"
-		"\\i \\i    \\\\[..]   matches one of a set          \\\\n     matches a newline           \\|\\i \\i\n"
-		"\\i \\i    \\\\+      matches 0 or more of the character which follows the \\\\+          \\|\\i \\i\n"
-		"\\i \\iSpecial replace sequences:                                                   \\|\\i \\i\n"
-		"\\i \\i    \\\\&      replaced with text which matched search string                   \\|\\i \\i\n"
-		"\\i \\i    \\\\0 - 9  replaced with text which matched Nth \\\\*, \\\\?, \\\\c, \\\\+, or \\\\[..]    \\|\\i \\i\n"
-		"\\i \\i    \\\\\\\\      replaced with \\\\               \\\\n     replaced with newline       \\|\\i \\i\n"
+		"\\i \\iSearch sequences:                                                            \\|\\i \\i\n"
+		"\\i \\i    \\\\^  \\\\$  matches beg./end of line       \\\\.     match any single char      \\|\\i \\i\n"
+		"\\i \\i    \\\\<  \\\\>  matches beg./end of word       \\\\!     match char or expression   \\|\\i \\i\n"
+		"\\i \\i    \\\\(  \\\\)  grouping                       \\\\|     match left or right         \\|\\i \\i\n"
+		"\\i \\i    \\\\[a-z]  matches one of a set                                             \\|\\i \\i\n"
+		"\\i \\i    \\\\{1,3}  match 1 - 3 occurrences        \\\\?     match 0 or 1 occurrence     \\|\\i \\i\n"
+		"\\i \\i    \\\\+      match 1 or more occurrences    \\\\*     match 0 or more occurrences \\|\\i \\i\n"
+		"\\i \\iReplace sequences:                                                           \\|\\i \\i\n"
+		"\\i \\i    \\\\&      replaced with entire match     \\\\1 - 9 replaced with Nth group   \\|\\i \\i\n"
+		"\\i \\i    \\\\u \\\\l   convert next to upper/lower    \\\\U \\\\L  case convert until \\\\E     \\|\\i \\i\n"
+		"}\n"
+		"{Escape sequences\n"
+		"\\i   Help Screen    \\|turn off with ^KH    prev. screen ^[,    next screen ^[.     \\i\n"
+		"\\i \\iEscape sequences: \\\\x{10fff}  Unicode code point   \\\\p{Ll}  Unicode category \\|\\i \\i\n"
+		"\\i \\i    \\\\i / \\\\I  Identifier start      \\\\t  tab          \\\\e  escape               \\|\\i \\i\n"
+		"\\i \\i    \\\\c / \\\\C  Identifier continue   \\\\n  newline      \\\\r  carriage return      \\|\\i \\i\n"
+		"\\i \\i    \\\\d / \\\\D  Digit / Not a digit   \\\\b  backspace  \\\\xFF  hex character        \\|\\i \\i\n"
+		"\\i \\i    \\\\w / \\\\W  Word / Not a word     \\\\a  alert      \\\\377  octal character      \\|\\i \\i\n"
+		"\\i \\i    \\\\s / \\\\S  Space / Not a space   \\\\f  formfeed     \\\\\\\\  backslash            \\|\\i \\i\n"
 		"}\n"
 		"{SearchOptions\n"
 		"\\i   Help Screen    \\|turn off with ^KH    prev. screen ^[,    next screen ^[.     \\i\n"
 		"\\i \\iSearch options:                                                              \\|\\i \\i\n"
-		"\\i \\i   r Replace                                                                 \\|\\i \\i\n"
-		"\\i \\i   k Restrict search to highlighted block, which can be rectangular          \\|\\i \\i\n"
-		"\\i \\i   b Search backward instead of forward                                      \\|\\i \\i\n"
-		"\\i \\i   i Ignore case                                                             \\|\\i \\i\n"
-		"\\i \\i   a Search across all loaded files                                          \\|\\i \\i\n"
-		"\\i \\i   e Search across all files in Grep or Compile error list                   \\|\\i \\i\n"
-		"\\i \\i   w Wrap to beginning of file for this search                               \\|\\i \\i\n"
-		"\\i \\i   n Do not wrap to beginning of file for this search                        \\|\\i \\i\n"
-		"\\i \\i   nnn Perform exaclty nnn replacements                                      \\|\\i \\i\n"
+		"\\i \\i     r Replace      k Restrict search to highlighted block                   \\|\\i \\i\n"
+		"\\i \\i     i Ignore case  b Search backwards instead of forwards                   \\|\\i \\i\n"
+		"\\i \\i                    a Search across all loaded files                         \\|\\i \\i\n"
+		"\\i \\i                    e Search across all files in Grep or Compile error list  \\|\\i \\i\n"
+		"\\i \\i w / n  Allow / prevent wrap to start of file                                \\|\\i \\i\n"
+		"\\i \\i x / y  Search text is standard format / JOE format regular expression       \\|\\i \\i\n"
+		"\\i \\i   nnn  Perform exactly nnn replacements                                     \\|\\i \\i\n"
 		"}\n"
 		"{Math\n"
 		"\\i   Help Screen    \\|turn off with ^KH    prev. screen ^[,    next screen ^[.     \\i\n"
@@ -237,7 +250,7 @@ unsigned char *builtins[]=
 		"\\i \\i\\| 112 | pqrs tuvw  xyz{ |}~\x7f    \xf0\xf1\xf2\xf3 \xf4\xf5\xf6\xf7  \xf8\xf9\xfa\xfb \xfc\xfd\xfe\xff | 240 \\|\\i \\i\n"
 		"}\n"
 		":windows		Bindings common to all windows\n"
-		"type		^@ TO \xff		Type a character\n"
+		"type		U+0 TO U+10FFFF\n"
 		"abort		^C		Abort window\n"
 		"abort		^K Q\n"
 		"abort		^K ^Q\n"
@@ -299,10 +312,10 @@ unsigned char *builtins[]=
 		":def compile mwind!,mfit!,querysave,query,scratch,\"* Build Log *\",rtn,bof,markb,eof,\" \",markk,blkdel,build\n"
 		":def grep_find mwind!,mfit!,scratch,\"* Grep Log *\",rtn,bof,markb,eof,\" \",markk,blkdel,grep\n"
 		":def man scratch,\"* Man Page *\",rtn,bof,markb,eof,\" \",markk,blkdel,\" \",ltarw,run,\"man -P cat -S 2:3 \"\n"
-		":def shell1 scratch,\"* Shell 1 *\",rtn,vtbknd!,eof\n"
-		":def shell2 scratch,\"* Shell 2 *\",rtn,vtbknd!,eof\n"
-		":def shell3 scratch,\"* Shell 3 *\",rtn,vtbknd!,eof\n"
-		":def shell4 scratch,\"* Shell 4 *\",rtn,vtbknd!,eof\n"
+		":def shell1 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 1 *\",rtn,vtbknd!,eof\n"
+		":def shell2 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 2 *\",rtn,vtbknd!,eof\n"
+		":def shell3 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 3 *\",rtn,vtbknd!,eof\n"
+		":def shell4 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 4 *\",rtn,vtbknd!,eof\n"
 		":def shell_clear psh,bof,markb,eof,markk,blkdel\n"
 		":def shell_parse parserr\n"
 		":def shell_gparse gparse\n"
@@ -318,6 +331,7 @@ unsigned char *builtins[]=
 		":def shell_markk markk\n"
 		"paste			^[ ] 5 2 ;		Base64 paste (obsolete)\n"
 		"brpaste			^[ [ 2 0 0 ~		Bracketed paste\n"
+		"brpaste_done		^[ [ 2 0 1 ~		Bracketed paste done\n"
 		"insc			^[ [ 2 ~\n"
 		"insc			^[ [ L			SCO\n"
 		"delch			^[ [ 3 ~\n"
@@ -509,6 +523,7 @@ unsigned char *builtins[]=
 		"splitw		^K ^O\n"
 		"splitw		^K o\n"
 		"stat		^K SP		Show status\n"
+		"stat		^K ^@		Show status\n"
 		"tag		^K ;		Tags file search\n"
 		"tomatch		^G		Goto matching parenthasis\n"
 		"undo		^_		Undo changes\n"
@@ -603,9 +618,9 @@ unsigned char *builtins[]=
 		":query			Single-key query window\n"
 		":inherit windows\n"
 		":querya			Single-key query window for quoting\n"
-		"type		^@ TO \xff\n"
+		"type		U+0 TO U+10FFFF\n"
 		":querysr		Search & replace query window\n"
-		"type		^@ TO \xff\n"
+		"type		U+0 TO U+10FFFF\n"
 		":shell			Input to shell windows\n"
 		":inherit main\n"
 		"\"\x03\"		^C		Abort\n"
@@ -632,165 +647,101 @@ unsigned char *builtins[]=
 		"\"\x08\"		^H		Backspace\n"
 		"\"\x0d\"		^M		Return\n"
 		"\"\x7f\"		^?		Backspace\n"
-,	USTR "ftyperc",
-		USTR "*\n"
-		"-wordwrap\n"
+,	"ftyperc",
+		"[text]\n"
+		"*\n"
+		"*.txt\n"
+		"[not_text]\n"
 		"*.*\n"
+		"-wordwrap\n"
+		"[diff]\n"
 		"*\n"
+		"+\\? [A-Za-z]\n"
+		"+Index: [A-Za-z]\n"
+		"+Only in\n"
+		"+Nur in\n"
 		"+--- \n"
-		"-syntax diff\n"
-		"*\n"
-		"+*** \n"
-		"-syntax diff\n"
-		"*\n"
-		"+\\[1-9]\\+\\[0-9]\\[cda]\n"
-		"-syntax diff\n"
+		"+\\*\\*\\* \n"
+		"+[1-9][0-9]*[cda]\n"
 		"*.diff\n"
-		"-syntax diff\n"
 		"*.patch\n"
 		"-syntax diff\n"
-		"*\n"
-		"+Only in\n"
-		"-syntax diff\n"
-		"*\n"
-		"+Nur in\n"
-		"-syntax diff\n"
-		"*\n"
-		"+Index: \\[A-Za-z]\n"
-		"-syntax diff\n"
-		"*\n"
-		"+? \\[A-Za-z]\n"
-		"-syntax diff\n"
+		"[troff]\n"
 		"*.1\n"
-		"-wordwrap\n"
-		"-syntax troff\n"
 		"*.1.in\n"
 		"-wordwrap\n"
 		"-syntax troff\n"
+		"[tex]\n"
 		"*.tex\n"
-		"-syntax tex\n"
-		"-tex_comment\n"
-		"-no_double_quoted\n"
 		"*.sty\n"
 		"-syntax tex\n"
 		"-tex_comment\n"
 		"-no_double_quoted\n"
-		"*.txt\n"
 		"-wordwrap\n"
+		"[email]\n"
 		"*.article*\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*.followup\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*.letter\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*tmp/nn.*\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*tmp/Re*\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*tmp/snd.*\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*tmp/dmt*\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*tmp/pico.*\n"
+		"*tmp/mutt-*\n"
 		"-wordwrap\n"
 		"-syntax mail\n"
+		"[avr]\n"
+		"*.avr\n"
+		"*.inc\n"
+		"-syntax avr\n"
+		"-autoindent\n"
+		"-wordwrap\n"
+		"-tab 8\n"
+		"[assembly]\n"
 		"*.asm\n"
-		"-wordwrap\n"
-		"-syntax asm\n"
 		"*.s\n"
-		"-wordwrap\n"
-		"-syntax asm\n"
 		"*.S\n"
 		"-wordwrap\n"
 		"-syntax asm\n"
+		"[mason]\n"
 		"*.mas\n"
 		"-autoindent\n"
 		"-syntax mason\n"
 		"-smarthome\n"
 		"-smartbacks\n"
+		"[sml]\n"
 		"*.sml\n"
 		"-autoindent\n"
 		"-syntax sml\n"
 		"-smarthome\n"
 		"-smartbacks\n"
 		"-istep 2\n"
+		"[ocaml]\n"
 		"*.ml\n"
-		"-autoindent\n"
-		"-syntax ocaml\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-istep 2\n"
 		"*.mli\n"
 		"-autoindent\n"
 		"-syntax ocaml\n"
 		"-smarthome\n"
 		"-smartbacks\n"
 		"-istep 2\n"
+		"[ruby]\n"
 		"*.rb\n"
-		"-autoindent\n"
-		"-pound_comment\n"
-		"-syntax ruby\n"
-		"-text_delimiters do=end\n"
 		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/ ]ruby\\>\n"
-		"-autoindent\n"
-		"-pound_comment\n"
-		"-syntax ruby\n"
-		"-text_delimiters do=end\n"
+		"+#![ 	]*[^\\n]*ruby[0-9.]*\\>\n"
 		"*.gemspec\n"
-		"-autoindent\n"
-		"-pound_comment\n"
-		"-syntax ruby\n"
-		"-text_delimiters do=end\n"
 		"*Gemfile\n"
-		"-autoindent\n"
-		"-pound_comment\n"
-		"-syntax ruby\n"
-		"-text_delimiters do=end\n"
 		"*Rakefile\n"
-		"-autoindent\n"
-		"-pound_comment\n"
-		"-syntax ruby\n"
-		"-text_delimiters do=end\n"
 		"*Guardfile\n"
-		"-autoindent\n"
-		"-pound_comment\n"
-		"-syntax ruby\n"
-		"-text_delimiters do=end\n"
 		"*.rabl\n"
 		"-autoindent\n"
 		"-pound_comment\n"
 		"-syntax ruby\n"
 		"-text_delimiters do=end\n"
+		"[perl]\n"
 		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]\\[/ ]perl\\>\n"
-		"-autoindent\n"
-		"-syntax perl\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-pound_comment\n"
-		"-single_quoted\n"
+		"+#![ 	]*[^\\n]*perl[0-9.]*\\>\n"
 		"*.pl\n"
-		"-autoindent\n"
-		"-syntax perl\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-pound_comment\n"
-		"-single_quoted\n"
 		"*.pm\n"
-		"-autoindent\n"
-		"-syntax perl\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-pound_comment\n"
-		"-single_quoted\n"
 		"*.t\n"
 		"-autoindent\n"
 		"-syntax perl\n"
@@ -798,6 +749,7 @@ unsigned char *builtins[]=
 		"-smartbacks\n"
 		"-pound_comment\n"
 		"-single_quoted\n"
+		"[sql]\n"
 		"*.sql\n"
 		"-autoindent\n"
 		"-syntax sql\n"
@@ -810,6 +762,7 @@ unsigned char *builtins[]=
 		"-cpp_comment\n"
 		"-text_delimiters BEGIN|Begin|begin=END|End|end\n"
 		"-cpara >#!;*/%\n"
+		"[awk]\n"
 		"*.awk\n"
 		"-autoindent\n"
 		"-syntax awk\n"
@@ -817,6 +770,7 @@ unsigned char *builtins[]=
 		"-smartbacks\n"
 		"-purify\n"
 		"-pound_comment\n"
+		"[yacc]\n"
 		"*.y\n"
 		"-autoindent\n"
 		"-syntax c\n"
@@ -829,18 +783,8 @@ unsigned char *builtins[]=
 		"-cpp_comment\n"
 		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
 		"-cpara >#!;*/%\n"
+		"[lex]\n"
 		"*.l\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.lex\n"
 		"-autoindent\n"
 		"-syntax c\n"
@@ -853,15 +797,8 @@ unsigned char *builtins[]=
 		"-cpp_comment\n"
 		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
 		"-cpara >#!;*/%\n"
+		"[ada]\n"
 		"*.adb\n"
-		"-syntax ada\n"
-		"-autoindent\n"
-		"-istep 2\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-vhdl_comment\n"
-		"-text_delimiters declare|Declare|DECLARE|exception|Exception|EXCEPTION|if|If|IF|loop|Loop|LOOP|case|Case|CASE|package|Package|PACKAGE|procedure|Procedure|PROCEDURE|record|Record|RECORD|function|Function|FUNCTION=end|End|END\n"
 		"*.ads\n"
 		"-syntax ada\n"
 		"-autoindent\n"
@@ -871,111 +808,35 @@ unsigned char *builtins[]=
 		"-purify\n"
 		"-vhdl_comment\n"
 		"-text_delimiters declare|Declare|DECLARE|exception|Exception|EXCEPTION|if|If|IF|loop|Loop|LOOP|case|Case|CASE|package|Package|PACKAGE|procedure|Procedure|PROCEDURE|record|Record|RECORD|function|Function|FUNCTION=end|End|END\n"
+		"[cobol]\n"
 		"*.cbl\n"
-		"-syntax cobol\n"
 		"*.cob\n"
 		"-syntax cobol\n"
+		"[rexx]\n"
 		"*.rex\n"
-		"-syntax rexx\n"
-		"-autoindent\n"
 		"*.REX\n"
-		"-syntax rexx\n"
-		"-autoindent\n"
 		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]/rxx\\>\n"
+		"+#![ 	]*[^\\n]*rexx[0-9.]*\\>\n"
 		"-syntax rexx\n"
 		"-autoindent\n"
+		"[sed]\n"
 		"*.sed\n"
 		"-syntax sed\n"
+		"[postscript]\n"
 		"*.ps\n"
-		"-syntax ps\n"
 		"*.eps\n"
 		"-syntax ps\n"
+		"[c]\n"
 		"*.c\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.cpp\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.cc\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.c++\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.h\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.h.in\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.hpp\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
 		"*.h++\n"
+		"*.hh\n"
+		"*.mm\n"
 		"-autoindent\n"
 		"-syntax c\n"
 		"-highlighter_context\n"
@@ -987,28 +848,8 @@ unsigned char *builtins[]=
 		"-cpp_comment\n"
 		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
 		"-cpara >#!;*/%\n"
-		"*.hh\n"
-		"-autoindent\n"
-		"-syntax c\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
-		"-cpara >#!;*/%\n"
+		"[verilog]\n"
 		"*.v\n"
-		"-autoindent\n"
-		"-syntax verilog\n"
-		"-istep 2\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-c_comment\n"
-		"-cpp_comment\n"
-		"-text_delimiters `ifdef|`ifndef=`else=`endif:begin=end:case|casex|casez=endcase:function=endfunction:module=endmodule:task=endtask:attribute=endattribute:primitive=endprimitive:table=endtable\n"
-		"-cpara >#!;*/%\n"
 		"*.vh\n"
 		"-autoindent\n"
 		"-syntax verilog\n"
@@ -1020,6 +861,7 @@ unsigned char *builtins[]=
 		"-cpp_comment\n"
 		"-text_delimiters `ifdef|`ifndef=`else=`endif:begin=end:case|casex|casez=endcase:function=endfunction:module=endmodule:task=endtask:attribute=endattribute:primitive=endprimitive:table=endtable\n"
 		"-cpara >#!;*/%\n"
+		"[vhdl]\n"
 		"*.vhd\n"
 		"-autoindent\n"
 		"-syntax vhdl\n"
@@ -1030,29 +872,36 @@ unsigned char *builtins[]=
 		"-vhdl_comment\n"
 		"-text_delimiters entity|ENTITY|if|IF|component|COMPONENT|loop|LOOP|configuration|CONFIGURATION|units|UNITS|record|RECORD|case|CASE|function|FUNCTION|package|PACKAGE|architecture|ARCHITECTURE|block|BLOCK|process|PROCESS|generate|GENERATE=end|END\n"
 		"-cpara >#!;*-/%\n"
+		"[xml]\n"
 		"*.xml\n"
-		"-autoindent\n"
-		"-syntax xml\n"
 		"*.xsd\n"
-		"-autoindent\n"
-		"-syntax xml\n"
 		"*\n"
-		"+<\\?xml\n"
+		"+<.xml\n"
+		"*.jnlp\n"
+		"*.resx\n"
+		"*build.xml\n"
 		"-autoindent\n"
 		"-syntax xml\n"
+		"[plist]\n"
 		"*.plist\n"
 		"-autoindent\n"
 		"-syntax xml\n"
+		"[html]\n"
 		"*.htm\n"
-		"-wordwrap\n"
-		"-autoindent\n"
-		"-syntax html\n"
 		"*.html\n"
 		"-wordwrap\n"
 		"-autoindent\n"
+		"-smarthome\n"
+		"-smartbacks\n"
 		"-syntax html\n"
+		"[css]\n"
 		"*.css\n"
+		"-wordwrap\n"
+		"-autoindent\n"
+		"-smarthome\n"
+		"-smartbacks\n"
 		"-syntax css\n"
+		"[java]\n"
 		"*.java\n"
 		"-autoindent\n"
 		"-syntax java\n"
@@ -1062,10 +911,10 @@ unsigned char *builtins[]=
 		"-cpp_comment\n"
 		"-purify\n"
 		"-cpara >#!;*/%\n"
-		"*.jnlp\n"
-		"-autoindent\n"
-		"-syntax xml\n"
+		"[javascript]\n"
 		"*.js\n"
+		"*\n"
+		"+#![ 	]*[^\\n]*nodejs[0-9.]*\\>\n"
 		"-autoindent\n"
 		"-syntax js\n"
 		"-smarthome\n"
@@ -1074,11 +923,13 @@ unsigned char *builtins[]=
 		"-c_comment\n"
 		"-cpp_comment\n"
 		"-cpara >#!;*/%\n"
+		"[json]\n"
 		"*.json\n"
 		"-autoindent\n"
 		"-syntax json\n"
 		"-smarthome\n"
 		"-smartbacks\n"
+		"[typescript]\n"
 		"*.ts\n"
 		"-autoindent\n"
 		"-syntax typescript\n"
@@ -1088,22 +939,16 @@ unsigned char *builtins[]=
 		"-c_comment\n"
 		"-cpp_comment\n"
 		"-cpara >#!;*/%\n"
+		"[php]\n"
 		"*.php\n"
 		"-autoindent\n"
 		"-syntax php\n"
 		"-smarthome\n"
 		"-smartbacks\n"
 		"-purify\n"
+		"[python]\n"
 		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]\\+\\[ 	]python\\+\\[0-9]\\>\n"
-		"-autoindent\n"
-		"-syntax python\n"
-		"-highlighter_context\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-single_quoted\n"
-		"-pound_comment\n"
+		"+#![ 	]*[^\\n]*python[0-9.]*\\>\n"
 		"*.py\n"
 		"-autoindent\n"
 		"-syntax python\n"
@@ -1113,138 +958,51 @@ unsigned char *builtins[]=
 		"-purify\n"
 		"-single_quoted\n"
 		"-pound_comment\n"
+		"[csh]\n"
 		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]/csh\\>\n"
-		"-autoindent\n"
-		"-syntax csh\n"
-		"-pound_comment\n"
+		"+#![ 	]*[^\\n 	]*/csh\\>\n"
+		"+#![ 	]*[^\\n 	]*/tcsh\\>\n"
 		"*.csh\n"
-		"-autoindent\n"
-		"-syntax csh\n"
-		"-pound_comment\n"
 		"*.login\n"
-		"-autoindent\n"
-		"-syntax csh\n"
-		"-pound_comment\n"
 		"*.logout\n"
-		"-autoindent\n"
-		"-syntax csh\n"
-		"-pound_comment\n"
-		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]/tcsh\\>\n"
-		"-autoindent\n"
-		"-syntax csh\n"
-		"-pound_comment\n"
 		"*.tcsh\n"
-		"-autoindent\n"
-		"-syntax csh\n"
-		"-pound_comment\n"
 		"*.tcshrc\n"
 		"-autoindent\n"
 		"-syntax csh\n"
 		"-pound_comment\n"
+		"[sh]\n"
 		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]/sh\\>\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
+		"+#![ 	]*[^\\n 	]*/sh\\>\n"
+		"+#![ 	]*[^\\n 	]*/bash\\>\n"
 		"*.sh\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
 		"*profile\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
-		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]/bash\\>\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
 		"*.bash\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
 		"*.bashrc\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
 		"*.bash_login\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
 		"*.bash_logout\n"
 		"-autoindent\n"
 		"-syntax sh\n"
 		"-highlighter_context\n"
 		"-pound_comment\n"
 		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
+		"[lisp]\n"
 		"*.lisp\n"
-		"-autoindent\n"
-		"-syntax lisp\n"
-		"-semi_comment\n"
 		"*.lsp\n"
-		"-autoindent\n"
-		"-syntax lisp\n"
-		"-semi_comment\n"
 		"*.el\n"
 		"-autoindent\n"
 		"-syntax lisp\n"
 		"-semi_comment\n"
+		"[ksh]\n"
 		"*\n"
-		"+#!\\+\\[ 	]\\+\\[a-z/]/ksh\\>\n"
-		"-autoindent\n"
-		"-syntax sh\n"
-		"-highlighter_context\n"
-		"-pound_comment\n"
-		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
+		"+#![ 	]*[^\\n 	]*/ksh\\>\n"
 		"*.ksh\n"
 		"-autoindent\n"
 		"-syntax sh\n"
 		"-highlighter_context\n"
 		"-pound_comment\n"
 		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
-		"*Makefile\n"
-		"-autoindent\n"
-		"-syntax conf\n"
-		"-pound_comment\n"
-		"*makefile.in\n"
-		"-autoindent\n"
-		"-syntax conf\n"
-		"-pound_comment\n"
-		"*Makefile.in\n"
-		"-autoindent\n"
-		"-syntax conf\n"
-		"-pound_comment\n"
-		"*makefile\n"
-		"-autoindent\n"
-		"-syntax conf\n"
-		"-pound_comment\n"
-		"*AKEFILE\n"
-		"-autoindent\n"
-		"-syntax conf\n"
-		"-pound_comment\n"
+		"[pascal]\n"
 		"*.p\n"
-		"-autoindent\n"
-		"-syntax pascal\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-purify\n"
-		"-text_delimiters begin|BEGIN|record|RECORD|case|CASE=end|END:repeat|REPEAT=until|UNTIL\n"
 		"*.pas\n"
 		"-autoindent\n"
 		"-syntax pascal\n"
@@ -1252,18 +1010,14 @@ unsigned char *builtins[]=
 		"-smartbacks\n"
 		"-purify\n"
 		"-text_delimiters begin|BEGIN|record|RECORD|case|CASE=end|END:repeat|REPEAT=until|UNTIL\n"
+		"[fortran]\n"
 		"*.f\n"
-		"-autoindent\n"
-		"-syntax fortran\n"
 		"*.f90\n"
-		"-autoindent\n"
-		"-syntax fortran\n"
 		"*.for\n"
-		"-autoindent\n"
-		"-syntax fortran\n"
 		"*.FOR\n"
 		"-autoindent\n"
 		"-syntax fortran\n"
+		"[skill]\n"
 		"*.il\n"
 		"-semi_comment\n"
 		"-syntax skill\n"
@@ -1273,16 +1027,17 @@ unsigned char *builtins[]=
 		"-smartbacks\n"
 		"-smarthome\n"
 		"-indentfirst\n"
+		"[lua]\n"
 		"*.lua\n"
-		"-autoindent\n"
-		"-syntax lua\n"
 		"*\n"
-		"+#!\\+\\[     ]\\+\\[a-z/ ]lua\\>\n"
+		"+#![ 	]*[^\\n]*lua[0-9.]*\\>\n"
 		"-autoindent\n"
 		"-syntax lua\n"
+		"[tcl]\n"
 		"*.tcl\n"
 		"-autoindent\n"
 		"-syntax tcl\n"
+		"[go]\n"
 		"*.go\n"
 		"-autoindent\n"
 		"-syntax go\n"
@@ -1293,6 +1048,7 @@ unsigned char *builtins[]=
 		"-c_comment\n"
 		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
 		"-cpara >#!;*/%\n"
+		"[c#]\n"
 		"*.cs\n"
 		"-autoindent\n"
 		"-syntax csharp\n"
@@ -1301,113 +1057,88 @@ unsigned char *builtins[]=
 		"-purify\n"
 		"-cpara >#!;*/%\n"
 		"-text_delimiters #region=#endregion:#if=#elif=#else=#endif\n"
-		"*.resx\n"
-		"-autoindent\n"
-		"-syntax xml\n"
+		"[jsf]\n"
 		"*.jsf\n"
-		"-autoindent\n"
-		"-syntax jsf\n"
-		"-smarthome\n"
 		"*.jsf.in\n"
 		"-autoindent\n"
 		"-syntax jsf\n"
 		"-smarthome\n"
+		"[m4]\n"
 		"*.ac\n"
-		"-autoindent\n"
-		"-syntax m4\n"
 		"*.m4\n"
 		"-autoindent\n"
 		"-syntax m4\n"
+		"[matlab]\n"
 		"*.m\n"
 		"-wordwrap\n"
 		"-syntax matlab\n"
+		"[conf]\n"
 		"*.am\n"
-		"-autoindent\n"
-		"-syntax conf\n"
-		"*tmp/mutt-*\n"
-		"-wordwrap\n"
-		"-syntax mail\n"
 		"*rc\n"
-		"-syntax conf\n"
-		"*.bat\n"
+		"*Makefile\n"
+		"*makefile.in\n"
+		"*Makefile.in\n"
+		"*makefile\n"
+		"*AKEFILE\n"
 		"-autoindent\n"
-		"-syntax batch\n"
+		"-syntax conf\n"
+		"-pound_comment\n"
+		"[bat]\n"
+		"*.bat\n"
 		"*.cmd\n"
 		"-autoindent\n"
 		"-syntax batch\n"
+		"[joerc]\n"
 		"*ftyperc\n"
-		"-syntax joerc\n"
 		"*joerc\n"
-		"-syntax joerc\n"
 		"*jmacsrc\n"
-		"-syntax joerc\n"
 		"*jstarrc\n"
-		"-syntax joerc\n"
 		"*rjoerc\n"
-		"-syntax joerc\n"
 		"*jpicorc\n"
-		"-syntax joerc\n"
 		"*joerc.in\n"
-		"-syntax joerc\n"
 		"*jmacsrc.in\n"
-		"-syntax joerc\n"
 		"*jstarrc.in\n"
-		"-syntax joerc\n"
 		"*rjoerc.in\n"
-		"-syntax joerc\n"
 		"*jpicorc.in\n"
 		"-syntax joerc\n"
+		"[markdown]\n"
 		"*.md\n"
 		"-syntax md\n"
 		"-wordwrap\n"
+		"[yaml]\n"
 		"*.yml\n"
-		"-autoindent\n"
-		"-smarthome\n"
-		"-smartbacks\n"
-		"-pound_comment\n"
-		"-syntax yaml\n"
 		"*.yaml\n"
 		"-autoindent\n"
 		"-smarthome\n"
 		"-smartbacks\n"
 		"-pound_comment\n"
 		"-syntax yaml\n"
+		"[debian]\n"
 		"*apt/sources.list\n"
-		"-pound_comment\n"
-		"-syntax debian\n"
 		"*sources.list.*\n"
 		"-pound_comment\n"
 		"-syntax debian\n"
-		"*build.xml\n"
-		"-syntax ant\n"
+		"[java_properties]\n"
 		"*.properties\n"
 		"-encoding iso-8859-1\n"
 		"-pound_comment\n"
 		"-syntax properties\n"
+		"[ini]\n"
 		"*.ini\n"
-		"-syntax ini\n"
-		"-pound_comment\n"
-		"-semi_comment\n"
 		"*.inf\n"
-		"-syntax ini\n"
-		"-pound_comment\n"
-		"-semi_comment\n"
 		"*hgrc\n"
-		"-pound_comment\n"
-		"-semi_comment\n"
-		"-syntax ini\n"
 		"*.gitconfig\n"
-		"-pound_comment\n"
-		"-semi_comment\n"
-		"-syntax ini\n"
 		"*.git/config\n"
+		"-syntax ini\n"
 		"-pound_comment\n"
 		"-semi_comment\n"
-		"-syntax ini\n"
+		"[erb]\n"
 		"*.erb\n"
 		"-syntax erb\n"
+		"[erb_html]\n"
 		"*.html.erb\n"
 		"-syntax htmlerb\n"
+		"[haml]\n"
 		"*.haml\n"
 		"-syntax haml\n"
 		"-istep 2\n"
@@ -1415,55 +1146,127 @@ unsigned char *builtins[]=
 		"-autoindent\n"
 		"-smarthome\n"
 		"-smartbacks\n"
+		"[erlang]\n"
 		"*.erl\n"
-		"-syntax erlang\n"
-		"-autoindent\n"
 		"*.eterm\n"
-		"-syntax erlang\n"
-		"-autoindent\n"
 		"*.app.src\n"
-		"-syntax erlang\n"
-		"-autoindent\n"
 		"*rebar.config\n"
-		"-syntax erlang\n"
-		"-autoindent\n"
 		"*.hrl\n"
-		"-syntax erlang\n"
-		"-autoindent\n"
 		"*.erlang\n"
 		"-syntax erlang\n"
 		"-autoindent\n"
+		"[sieve]\n"
 		"*.sieve\n"
 		"-syntax sieve\n"
 		"-autoindent\n"
+		"[puppet]\n"
 		"*.pp\n"
 		"-syntax puppet\n"
 		"-pound_comment\n"
 		"-autoindent\n"
+		"[iptables]\n"
 		"*\n"
 		"+# Generated by iptables-save\n"
-		"-syntax iptables\n"
-		"-pound_comment\n"
-		"*\n"
 		"+# Generated by ip6tables-save\n"
 		"-syntax iptables\n"
 		"-pound_comment\n"
+		"[powershell]\n"
 		"*.ps1\n"
 		"-syntax powershell\n"
+		"[git-commit]\n"
 		"*COMMIT_EDITMSG\n"
 		"-syntax git-commit\n"
+		"[elixir]\n"
 		"*.ex\n"
-		"-autoindent\n"
-		"-syntax elixir\n"
-		"-pound_comment\n"
-		"-text_delimiters do=end\n"
 		"*.exs\n"
 		"-autoindent\n"
 		"-syntax elixir\n"
 		"-pound_comment\n"
 		"-text_delimiters do=end\n"
-,	USTR "c.jsf",
-		USTR "\n"
+		"[r]\n"
+		"*.r\n"
+		"-autoindent\n"
+		"-syntax r\n"
+		"-pound_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"-single_quoted\n"
+		"[groovy]\n"
+		"*.gvy\n"
+		"*.groovy\n"
+		"*.gy\n"
+		"*.gsh\n"
+		"-autoindent\n"
+		"-syntax groovy\n"
+		"-c_comment\n"
+		"-cpp_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"-single_quoted\n"
+		"[clojure]\n"
+		"*.clj\n"
+		"*.cljs\n"
+		"*.cljc\n"
+		"*.edn\n"
+		"-autoindent\n"
+		"-syntax clojure\n"
+		"-semi_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"[rust]\n"
+		"*.rs\n"
+		"-autoindent\n"
+		"-syntax rust\n"
+		"-c_comment\n"
+		"-cpp_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"-single_quoted\n"
+		"[coffee]\n"
+		"*.coffee\n"
+		"-autoindent\n"
+		"-syntax coffee\n"
+		"-pound_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"-single_quoted\n"
+		"[scala]\n"
+		"*.scala\n"
+		"-autoindent\n"
+		"-syntax scala\n"
+		"-c_comment\n"
+		"-cpp_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"-single_quoted\n"
+		"[swift]\n"
+		"*.swift\n"
+		"-autoindent\n"
+		"-syntax swift\n"
+		"-c_comment\n"
+		"-cpp_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"-single_quoted\n"
+		"[d]\n"
+		"*.d\n"
+		"-autoindent\n"
+		"-syntax d\n"
+		"-c_comment\n"
+		"-cpp_comment\n"
+		"-smarthome\n"
+		"-smartbacks\n"
+		"-purify\n"
+		"-single_quoted\n"
+,	"c.jsf",
+		"\n"
 		"\n"
 		"\n"
 		"\n"
@@ -1541,11 +1344,11 @@ unsigned char *builtins[]=
 		":predef_ws Preproc\n"
 		"	*		prebad		recolor=-1\n"
 		"	\" \\t\"		predef_ws\n"
-		"	\"a-zA-Z0-9_\"	predef_ident	recolor=-1\n"
+		"	\"\\c\"		predef_ident	recolor=-1\n"
 		"\n"
 		":predef_ident Define\n"
 		"	*		idle		noeat\n"
-		"	\"a-zA-Z0-9_\"	predef_ident\n"
+		"	\"\\c\"		predef_ident\n"
 		"\n"
 		"\n"
 		":preproc Preproc\n"
@@ -1573,7 +1376,7 @@ unsigned char *builtins[]=
 		"	\".\"		maybe_float\n"
 		"	\"\\\"\"		string		recolor=-1\n"
 		"	\"'\"		char		recolor=-1\n"
-		"	\"a-zA-Z_\"	ident		buffer\n"
+		"	\"\\i\"		ident		buffer\n"
 		"	\"\\\\\"		outside_escape	recolor=-1\n"
 		"	\"{}\"		brace		recolor=-1\n"
 		"	\",:;=()><[]*&|!~+\\-%^\"	control		recolor=-1\n"
@@ -1772,7 +1575,7 @@ unsigned char *builtins[]=
 		"	\"wchar_t\"		type\n"
 		"	\"typeof\"		cppkw\n"
 		"done\n"
-		"	\"a-zA-Z0-9_\"	ident\n"
+		"	\"\\c\"		ident\n"
 		"\n"
 		":type Type\n"
 		"	*	idle	noeat\n"
